"""
This module contains all of the constants used in ``modern_urwid``.

Module variables:

- ``XML_NS`` - XML namespace for modern-urwid
- ``RESOURCE_CHAR`` - The character used to reference resources from ResourceHandler
- ``DEFAULT_STYLE`` - The default style for widgets
"""

XML_NS: str = "{https://github.com/Jackkillian/modern-urwid}"
RESOURCE_CHAR: str = "@"
DEFAULT_STYLE: dict[str, str] = {
    "color": "",
    "background": "",
    "monochrome": "",
    "color-adv": "",
    "background-adv": "",
}



from __future__ import annotations

import importlib
import string
from pathlib import Path
from types import ModuleType
from typing import Union

import urwid
from lxml import etree

from modern_urwid.resource_handler import ResourceHandler
from modern_urwid.xml_parser import XMLParser

from .builder import WidgetBuilder
from .constants import XML_NS
from .css_parser import CSSParser
from .exceptions import UnknownResource


class LayoutResourceHandler(ResourceHandler):
    """
    A base class for extending a layout's functionality.

    Reference properties from the base class (e.g. callbacks) with the ``@`` prefix.
    Reference properties from the data dictionary with brackets surrounding the key (e.g. ``"{user.name}"``).
    """

    def __init__(
        self,
        layout: Layout,
        palettes=[],
        widgets: list[type[WidgetBuilder]] = [],
        css_variables: dict[str, str] = {},
    ):
        self.layout = layout
        self.palettes = palettes
        self.widgets = widgets
        self.css_variables = css_variables
        self.data = {}

    def get_palettes(self):
        """Get custom palettes."""
        return self.palettes

    def get_resource(self, name):
        """Get a custom resource (typically referenced by ``@ResourceName`` in XML)."""
        if hasattr(self, name):
            return getattr(self, name)
        else:
            raise UnknownResource(f"Could not custom resource '@{name}'")

    def parse_string_template(self, template):
        variables = [
            field for _, field, _, _ in string.Formatter().parse(template) if field
        ]
        value = template
        for variable in variables:
            value = value.replace(f"{{{variable}}}", self._get_data_resource(variable))
        return value

    def _get_data_resource(self, attr):
        keys = attr.split(".")
        value = self.data
        for key in keys:
            if isinstance(value, dict):
                if key not in value:
                    raise ValueError(f"Could not find key '{key}' on {{{attr}}}")
                value = value.get(key)
            elif isinstance(value, ModuleType):
                if hasattr(value, key):
                    value = getattr(value, key)
                else:
                    raise ValueError(f"Could not find key '{key}' on {{{attr}}}")
        return value

    def parse_resources_tag(self, element: etree.Element):
        for child in element:
            if child.tag == f"{XML_NS}python":
                module_path = child.get("module")
                if module_path is None:
                    raise ValueError(
                        "Could not get attribute 'module' for mu:python element"
                    )

                alias = child.get("as")
                module = importlib.import_module(module_path)
                if alias:
                    self.data[alias] = module
                else:
                    keys = module_path.split(".")
                    target = self.data
                    for key in keys[:-1]:
                        if key not in target:
                            target[key] = {}
                            target = target[key]
                        else:
                            target = target[key]
                    target[keys[-1]] = module

    def get_widget_builder(self, tag: str) -> Union[type[WidgetBuilder], None]:
        cls_lower = tag.lower()
        for cls in self.widgets:
            if cls_lower == cls.__name__.lower():
                return cls

    def get_css_variables(self) -> dict[str, str]:
        return self.css_variables

    def on_load(self):
        """Called when loading the parent layout in :meth:`~modern_urwid.layout_manager.LayoutManager.register`."""
        return

    def on_enter(self):
        """Called when the parent layout is rendered on the mainloop with :meth:`~modern_urwid.layout_manager.LayoutManager.switch`."""
        pass

    def on_exit(self):
        """Called when the parent layout is removed from the mainloop with :meth:`~modern_urwid.layout_manager.LayoutManager.switch`."""
        pass


class Layout:
    """
    Create a UI layout from XML and CSS.

    This class is responsible for parsing XML and applying
    CSS styles to the created widgets. The root widget can
    be referenced with :meth:`get_root`.
    """

    def __init__(
        self,
        xml_path: Path,
        css_path: Union[Path, None] = None,
        resources_cls=LayoutResourceHandler,
        xml_dir=None,
        css_dir=None,
    ) -> None:
        self.resources = resources_cls(self)
        self.widget_map = {}

        # Handle path stuff
        self.css_path = css_path
        if css_path is not None:
            self.css_path = css_path
            if isinstance(css_dir, Path):
                self.css_path = css_dir / css_path
            self.css_dir = self.css_path.parent
        else:
            self.css_dir = None

        self.xml_path = xml_path
        if isinstance(xml_dir, Path):
            self.xml_path = xml_dir / xml_path
        self.xml_dir = self.xml_path.parent

    def register_widgets(self, widgets: list[type[WidgetBuilder]]):
        """Add custom widget builders for the XML parser. Note: do this before calling :meth:`load`."""
        self.resources.widgets.extend(widgets)

    def style_widget(self, widget: urwid.Widget, classes=[], id=None) -> urwid.AttrMap:
        """Style a widget according to any given classes or id it may have."""
        return self.xml_parser.style_widget(widget, classes, id)

    def load(self):
        """Parse the XML and CSS. Make sure to call :meth:`register_widgets` first if neccessary."""
        self.css_parser = CSSParser(self.css_path, self.resources.get_css_variables())
        self.xml_parser = XMLParser(self.xml_path, self.resources, self.css_parser)
        self.resources.on_load()
        return self

    def get_root(self) -> urwid.Widget:
        """Get the root XML widget that can be rendered."""
        return self.xml_parser.get_root()

    def get_palettes(self):
        """Get all palettes used in this layout."""
        if (palettes := self.resources.get_palettes()) is None:
            palettes = []
        return self.xml_parser.get_palettes() + palettes

    def get_widget_by_id(self, id) -> Union[urwid.Widget, None]:
        """Get a widget by its ``id`` attribute."""
        return self.xml_parser.get_widget_by_id(id)

    def on_enter(self):
        """Called when this is rendered on the mainloop with :meth:`~modern_urwid.layout_manager.LayoutManager.switch`."""
        self.resources.on_enter()

    def on_exit(self):
        """Called when this is removed from the mainloop with :meth:`~modern_urwid.layout_manager.LayoutManager.switch`."""
        self.resources.on_exit()



import importlib.resources
from pathlib import Path

import urwid

from modern_urwid import Layout, LayoutManager, LayoutResourceHandler, WidgetBuilder


def test_layout_loads():
    class CustomResources2(LayoutResourceHandler):
        pass

    class CustomResources(LayoutResourceHandler):
        def __init__(self, layout):
            self.Layout2Resources = CustomResources2(layout)
            super().__init__(
                layout,
                palettes=[
                    ("pb_empty", "white", "black"),
                    ("pb_full", "black", "light blue"),
                ],
                css_variables={"--my-var": "light gray"},
            )

        def quit_callback(self, w):
            raise urwid.ExitMainLoop()

        def on_edit_change(self, w: urwid.Edit, full_text):
            w.set_caption(f"Edit ({full_text}): ")

        def on_edit_postchange(self, w, text):
            widget = self.layout.get_widget_by_id("header_text")
            if isinstance(widget, urwid.Text):
                widget.set_text(text)

        def on_load(self):
            widget = self.layout.get_widget_by_id("dynamic")
            if isinstance(widget, urwid.ListBox):
                widget.body.extend(
                    [
                        self.layout.style_widget(
                            urwid.Button(f"Dynamic Button {i}"), id="root"
                        )
                        for i in range(10)
                    ]
                )

    manager = LayoutManager()

    @manager.register_widget()
    class CustomWidget(WidgetBuilder):
        def build(self, **kwargs):
            return urwid.Filler(urwid.Text("This is a custom widget!"))

    @manager.register_widget()
    class CustomWidgetFromXML(WidgetBuilder):
        def build(self, **kwargs):
            parser = self.render_from_xml(
                Path(importlib.resources.files("tests") / "resources" / "widget.xml"),
                css_path=Path(
                    importlib.resources.files("tests") / "resources" / "widget.css"
                ),
            )
            manager.register_palette(parser.get_palettes())
            return parser.get_root()

    manager.register(
        "layout",
        layout := Layout(
            Path(importlib.resources.files("tests") / "resources" / "layout.xml"),
            Path(importlib.resources.files("tests") / "resources" / "styles.css"),
            CustomResources,
        ),
    )

    assert isinstance(layout.get_root(), urwid.AttrMap)
    assert isinstance(layout.get_root().base_widget, urwid.Pile)

    screen: urwid.display.raw.Screen = manager.get_loop().screen
    screen.set_terminal_properties(2**24)

    manager.switch("layout")
    manager.run()

    # loop.start()
    # loop.screen.clear()
    # loop.draw_screen()

    # time.sleep(10)



from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from .builder import WidgetBuilder

from lxml import etree

from .exceptions import InvalidTemplate, UnknownResource


class ResourceHandler:
    def get_resource(self, name: str) -> Any:
        """Load a given resource, referenced by ``<tag attribute="@MyResource" />``"""
        raise UnknownResource(
            f"Could not get resource '@{name}' - get_resource() not implemented"
        )

    def parse_string_template(self, template) -> str:
        """Parse a given string template, e.g. ``"User: {user.id}"``"""
        raise InvalidTemplate(
            f"Could not parse template '{{{template}}}' - parse_string_template() not implemented"
        )

    def parse_resources_tag(self, element: etree.Element) -> None:
        """Parse the ``<mu:resources />`` tag"""
        raise ValueError(
            f"Could not parse element: {element} - load_resources_from_tag() not implemented"
        )

    def get_widget_builder(self, tag: str) -> Union[type["WidgetBuilder"], None]:
        """Get a widget builder for a custom widget tag."""
        return None

    def get_css_variables(self) -> dict[str, str]:
        """Get custom CSS variables, used to override stylesheet variables."""
        return {}



from .builder import WidgetBuilder
from .constants import RESOURCE_CHAR, XML_NS
from .exceptions import InvalidTemplate, UnknownResource
from .layout import Layout, LayoutResourceHandler
from .layout_manager import LayoutManager

__all__ = [
    "Layout",
    "LayoutResourceHandler",
    "LayoutManager",
    "WidgetBuilder",
    "UnknownResource",
    "InvalidTemplate",
    "XML_NS",
    "RESOURCE_CHAR",
]



import sys
from importlib.metadata import version as pkg_version
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[2] / "src"))


# -- Project information

project = "modern-urwid"
copyright = "2025, Jack Freund"
author = "Jack Freund"

release = pkg_version("modern-urwid")
version = ".".join(release.split(".")[:2])

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "myst_parser",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"

# -- Options for EPUB output
epub_show_urls = "footnote"


# -- Autosummary options
autosummary_generate = True

autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "show-inheritance": True,
}

# -- MyST Config
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]



from __future__ import annotations

import inspect
from pathlib import Path
from typing import Callable, Literal, Union

import urwid
from dict_hash import md5
from lxml import etree

from .constants import DEFAULT_STYLE, RESOURCE_CHAR, XML_NS
from .css_parser import CSSParser
from .resource_handler import ResourceHandler
from .wrapper import FilteredWrapper


def find_urwid_class(tag: str):
    tag = tag.lower()
    for name, cls in inspect.getmembers(urwid, inspect.isclass):
        if name.lower() == tag:
            return cls
    return None


def create_text_widget(cls, el, **kw):
    if el.text and el.text.strip():
        return cls(
            el.text, **kw
        )  # can't do .strip because it'll  remove the space if doing something like 'Name: '
    else:
        return cls(**kw)


class XMLParser:
    def __init__(
        self,
        xml_path: Path,
        resources: ResourceHandler = ResourceHandler(),
        css_parser: CSSParser = CSSParser(None),
        xml_dir: Union[Path, None] = None,
    ) -> None:
        self.widget_map = {}
        self.styles = {}
        self.xml_path = xml_path
        self.css_parser = css_parser
        self.resources = resources

        self.xml_path = xml_path
        if isinstance(xml_dir, Path):
            self.xml_path = xml_dir / xml_path
        self.xml_dir = self.xml_path.parent

        # Load the xml
        xml = self.xml_path.read_text()
        root = self.parse_element(
            FilteredWrapper.from_html_root(etree.fromstring(xml)),
            DEFAULT_STYLE,
        )
        if not isinstance(root, urwid.Widget):
            raise ValueError(f"Got {root} instead of Widget for root")
        else:
            self.root = root

    def get_root(self):
        return self.root

    def get_palettes(self):
        return [(hash, *style.values()) for hash, style in self.styles.items()]

    def parse_attrs(self, kwargs: dict):
        mu = {}
        normal = {}
        for k, v in kwargs.items():
            target = normal
            if k.startswith(XML_NS):
                k = k[len(XML_NS) :]
                target = mu
            if isinstance(v, str):
                if v.isdigit():
                    target[k] = int(v)
                elif v.startswith(RESOURCE_CHAR):
                    target[k] = self.resources.get_resource(v[len(RESOURCE_CHAR) :])
                elif v == "False":
                    target[k] = False
                elif v == "True":
                    target[k] = True
                else:
                    target[k] = self.resources.parse_string_template(v)
        return mu, normal

    def parse_element(
        self,
        wrapper: FilteredWrapper,
        root_palette: dict,
        child_class: Union[str, None] = None,
    ) -> Union[
        urwid.Widget,
        tuple[int, urwid.Widget],
        tuple[Literal["weight"], int, urwid.Widget],
    ]:
        if child_class is not None:
            wrapper.classes |= {child_class}

        element = wrapper.etree_element
        tag = element.tag

        # Parse attributes
        mu_kwargs, kwargs = self.parse_attrs(element.attrib)
        clazz = kwargs.pop("class", None)
        id = kwargs.pop("id", None)
        child_class = mu_kwargs.get("child_class")
        height = mu_kwargs.get("height")
        weight = mu_kwargs.get("weight")

        # Parse children
        signals = {}
        children = list(wrapper.iter_children())
        for child in wrapper.iter_mu_children():
            el = child.etree_element
            if el.tag == f"{XML_NS}signal":
                signal_name = el.get("name")
                signals[signal_name] = self.parse_attrs(el.attrib)
            elif el.tag == f"{XML_NS}resources":
                self.resources.parse_resources_tag(el)
            else:
                children.append(child)

        # Apply styling
        style, pseudos = self.css_parser.get_styles(root_palette, wrapper)

        normal_hash = md5(style)
        if normal_hash not in self.styles:
            self.styles[normal_hash] = style

        focus_hash = None
        if (
            "focus" in pseudos
            and (focus_hash := md5(pseudos["focus"])) not in self.styles
        ):
            self.styles[focus_hash] = {**style.copy(), **pseudos["focus"]}

        if constructor := self.get_widget_constructor(tag):
            widget = constructor(
                element,
                [self.parse_element(child, style, child_class) for child in children],
                **kwargs,
            )
        else:
            return urwid.Filler(urwid.Text(f"Unknown tag: {tag}"))

        if id is not None:
            if id in self.widget_map:
                raise ValueError(f"Cannot duplicate IDs: {id}")
            else:
                self.widget_map[id] = widget

        for name, attrs in signals.items():
            urwid.connect_signal(
                widget, name, attrs[1].get("callback"), attrs[1].get("user_arg")
            )

        if height is not None:
            return (height, urwid.AttrMap(widget, normal_hash, focus_hash))
        elif weight is not None:
            return (
                "weight",
                weight,
                urwid.AttrMap(widget, normal_hash, focus_hash),
            )

        return urwid.AttrMap(widget, normal_hash, focus_hash)

    def style_widget(self, widget: urwid.Widget, classes=[], id=[]) -> urwid.AttrMap:
        style, pseudos = self.css_parser.get_styles_by_attr(DEFAULT_STYLE, classes, id)

        normal_hash = md5(style)
        if normal_hash not in self.styles:
            self.styles[normal_hash] = style

        focus_hash = None
        if (
            "focus" in pseudos
            and (focus_hash := md5(pseudos["focus"])) not in self.styles
        ):
            self.styles[focus_hash] = {**style.copy(), **pseudos["focus"]}

        return urwid.AttrMap(widget, normal_hash, focus_hash)

    def get_widget_constructor(
        self, tag
    ) -> Union[
        Callable[
            [etree.Element, list[Union[urwid.Widget, urwid.WidgetContainerMixin]]],
            Union[urwid.Widget, urwid.WidgetContainerMixin],
        ],
        None,
    ]:
        if tag == f"{XML_NS}layout":

            def fun(el, children, **kw):
                parser = XMLParser(
                    **{
                        "xml_dir": self.xml_dir,
                        "resources": self.resources,
                        "css_parser": CSSParser(
                            self.css_parser.path.parent / kw.get("css_path", "")
                        )
                        if "css_path" in kw
                        else self.css_parser,
                        **{k: v for k, v in kw.items() if k != "css_path"},
                    }
                )
                self.styles.update(parser.styles)
                return parser.get_root()

            return fun

        if cls := self.resources.get_widget_builder(tag.lower()):
            return lambda el, children, **kw: cls(el, children).build(**kw)

        if (cls := find_urwid_class(tag)) is None:
            return None

        if issubclass(
            cls,
            urwid.WidgetContainerMixin,
        ):
            return lambda el, children, **kw: cls(children, **kw)
        elif issubclass(
            cls,
            urwid.WidgetDecoration,
        ):
            return lambda el, children, **kw: cls(children[0], **kw)
        elif issubclass(cls, urwid.Widget):
            return lambda el, children, **kw: create_text_widget(cls, el, **kw)
        else:
            return None

    def get_widget_by_id(self, id) -> Union[urwid.Widget, None]:
        return self.widget_map.get(id)



import cssselect2

from .constants import XML_NS


class FilteredWrapper(cssselect2.ElementWrapper):
    def iter_mu_children(self):
        child = None
        for i, etree_child in enumerate(self.etree_children):
            if etree_child.tag.startswith(XML_NS):
                child = type(self)(
                    etree_child,
                    parent=self,
                    index=i,
                    previous=child,
                    in_html_document=self.in_html_document,
                )
                yield child

    def iter_children(self):
        child = None
        for i, etree_child in enumerate(self.etree_children):
            if not etree_child.tag.startswith(XML_NS):
                child = type(self)(
                    etree_child,
                    parent=self,
                    index=i,
                    previous=child,
                    in_html_document=self.in_html_document,
                )
                yield child



from types import LambdaType
from typing import Union

import urwid

from .builder import WidgetBuilder
from .layout import Layout


class LayoutManager:
    """
    Manages multiple layouts and shared custom widgets and palettes
    between them.
    """

    def __init__(self, loop: Union[urwid.MainLoop, None] = None):
        if loop is None:
            self.loop = urwid.MainLoop(urwid.Text(""))
        else:
            self.loop: urwid.MainLoop = loop
        self.layouts: dict[str, Layout] = {}
        self.current: Union[str, None] = None
        self.widgets: list[type[urwid.WidgetBuilder]] = []

    def register(self, name: str, layout: Layout):
        """Register a new :class:`~modern_urwid.layout.Layout`"""
        self.layouts[name] = layout
        layout.register_widgets(self.widgets)
        layout.load()
        self.loop.screen.register_palette(layout.get_palettes())

    def switch(self, name: str):
        """
        Switch to a different layout by name.

        Calls the new layout's :meth:`~modern_urwid.layout.Layout.on_enter` method, and the
        old layout's :meth:`~modern_urwid.layout.Layout.on_exit` method.
        """
        if self.current:
            self.layouts[self.current].on_exit()

        layout = self.layouts[name]
        layout.on_enter()
        self.loop.widget = layout.get_root()
        self.current = name

    def register_palette(self, palette):
        """Register a set of palette entries in the urwid :class:`~urwid.MainLoop`."""
        self.loop.screen.register_palette(palette)

    def register_widget(
        self, cls: Union[type[WidgetBuilder], None] = None
    ) -> LambdaType:
        """
        Register a custom widget builder.

        This can be used either as a **decorator** (``@manager.register_widget()``)
        or by directly passing a class (``manager.register_widget(MyCustomBuilder)``).
        """

        def decorator(cls: type[WidgetBuilder]):
            self.widgets.append(cls)
            return cls

        if cls:
            self.widgets.append(cls)

        return decorator

    def get_loop(self) -> urwid.MainLoop:
        """Get the urwid :class:`~urwid.MainLoop`"""
        return self.loop

    def run(self):
        """Run the urwid :class:`~urwid.MainLoop`"""
        self.loop.run()



from pathlib import Path
from re import L
from typing import Union

import cssselect2
import tinycss2
from tinycss2.ast import (
    Declaration,
    FunctionBlock,
    HashToken,
    IdentToken,
    LiteralToken,
    Node,
    WhitespaceToken,
)


def pop_pseudos_from_tokens(tokens):
    result = []
    pseudos = []
    i = 0
    while i < len(tokens):
        token = tokens[i]
        if token.type == "literal" and token.value == ":":
            if len(tokens) > i + 1 and tokens[i + 1].type == "ident":
                pseudos.append(tokens[i + 1].value)
                i += 2
                continue
        elif token.type == "function":
            pseudos.append(token.name)
        else:
            result.append(token)
        i += 1
    return result, pseudos


def split_tokens_by_comma(tokens):
    selectors = []
    current = []
    for token in tokens:
        if isinstance(token, WhitespaceToken):
            continue
        if token.type == "literal" and token.value == ",":
            selectors.append(current)
            current = []
        else:
            current.append(token)
    if current:
        selectors.append(current)
    return selectors


def get_props(tokens, variables):
    modified = [
        IdentToken(-1, -1, variables.get(token.arguments[0].value))
        if isinstance(token, FunctionBlock)
        else token
        for token in tokens
    ]
    decls: list[Declaration] = tinycss2.parse_declaration_list(
        modified, skip_comments=True, skip_whitespace=True
    )
    return {
        decl.name: "".join(
            [
                "#" + token.value if isinstance(token, HashToken) else token.value
                for token in decl.value
                if hasattr(token, "value")
            ]
        ).strip()
        for decl in decls
        if not decl.name.startswith("--")
    }


def get_tokens_value(tokens: list[Node]) -> str:
    value = ""
    for token in tokens:
        value += token.serialize()
    return value


def split_decl(tokens):
    result = []
    name = []
    value = []
    in_name = True
    for token in tokens:
        if token.type == "literal":
            if token.value == ":":
                in_name = False
            elif token.value == ";":
                in_name = True
                result.append((name.copy(), value.copy()))
                name.clear()
                value.clear()
        elif in_name and not token.type == "whitespace":
            name.append(token)
        elif not in_name:
            value.append(token)
    return result


class CSSParser:
    def __init__(
        self, path: Union[Path, None], variables_override: dict[str, str] = {}
    ):
        self.matcher = cssselect2.Matcher()

        if path is None:
            return

        if not path.exists():
            raise FileNotFoundError(f"Could not find stylesheet: {path} does not exist")
        elif not path.is_file():
            raise IsADirectoryError(f"Could not find stylesheet: {path} is a directory")

        self.path = path
        self.dir = path.parent

        css = path.read_text()
        rules: list[Node] = tinycss2.parse_stylesheet(
            css,
            skip_comments=True,
            skip_whitespace=True,
        )
        self.parse_rules(rules, variables_override)

    def parse_rules(self, rules, variables={}):
        self.pseudo_map = {}
        for rule in rules:
            if rule.type != "qualified-rule":
                continue

            for name, value in split_decl(rule.content):
                name = get_tokens_value(name)
                value = get_tokens_value(value).strip()
                if name.startswith("--"):
                    if name not in variables:
                        variables[name] = value

            element_selectors: list[list[Node]] = split_tokens_by_comma(rule.prelude)
            props = get_props(rule.content, variables)
            for selectors in element_selectors:
                compiled = cssselect2.compile_selector_list(selectors)

                selectors, pseudos = pop_pseudos_from_tokens(
                    selectors
                )  # NOTE: overwrites selectors
                sel_str = tinycss2.serialize(selectors)

                for item in compiled:
                    self.matcher.add_selector(item, (sel_str, props))

                for pseudo in pseudos:
                    if sel_str not in self.pseudo_map:
                        self.pseudo_map[sel_str] = {}
                    self.pseudo_map[sel_str][pseudo] = props

    def get_styles(self, default: dict[str, str], element: cssselect2.ElementWrapper):
        style = default.copy()
        pseudos = {}
        if matches := self.matcher.match(element):
            matches.sort()
            for match in matches:
                specificity, order, pseudo, payload = match
                sel_str, data = payload
                style.update(data)

                # Default to 8-bit colors if true colors are not defined
                if "color-adv" not in data:
                    style["color-adv"] = style["color"]

                if "background-adv" not in data:
                    style["background-adv"] = style["background"]

                if sel_str in self.pseudo_map:
                    pseudos = self.pseudo_map[sel_str]

        return style, pseudos

    def get_styles_by_attr(self, default, classes=[], id=None):
        style = default.copy()
        pseudos = {}
        if matches := self.match(classes, id):
            matches.sort()
            for match in matches:
                specificity, order, pseudo, payload = match
                sel_str, data = payload
                style.update(data)
                if sel_str in self.pseudo_map:
                    pseudos = self.pseudo_map[sel_str]
        return style, pseudos

    def match(self, classes=[], id: Union[str, None] = None):
        relevant_selectors = []

        if id is not None and id in self.matcher.id_selectors:
            for test, specificity, order, pseudo, payload in self.matcher.id_selectors[
                id
            ]:
                relevant_selectors.append((specificity, order, pseudo, payload))

        for class_name in classes:
            if class_name in self.matcher.class_selectors:
                for (
                    test,
                    specificity,
                    order,
                    pseudo,
                    payload,
                ) in self.matcher.class_selectors[class_name]:
                    relevant_selectors.append((specificity, order, pseudo, payload))

        relevant_selectors.sort()
        return relevant_selectors



from pathlib import Path
from typing import Union

import urwid
from lxml import etree

from .css_parser import CSSParser
from .resource_handler import ResourceHandler
from .xml_parser import XMLParser


class WidgetBuilder:
    """
    Utility class used to build custom widgets. Widgets can be made
    with Python or XML. Use the ``LayoutManager.register_widget()`` decorator to register.
    """

    def __init__(
        self,
        element: etree.Element,
        children: list[Union[urwid.Widget, urwid.WidgetContainerMixin]] = [],
    ):
        self.element = element
        self.children = children

    def build(self, **kwargs) -> urwid.Widget:
        """Build the widget. When overriding, use the ``self.manager``, ``self.element``, ``self.children`` if applicable."""
        return urwid.Widget(**kwargs)

    def render_from_xml(
        self,
        xml_path: Path,
        resource_handler=ResourceHandler(),
        css_path: Union[Path, None] = None,
    ) -> XMLParser:
        """Render a widget from XML. Note: ``XMLParser.styles`` will need to be registered in urwid palettes."""
        return XMLParser(xml_path, resource_handler, CSSParser(css_path))
